@@ -33,7 +33,7 @@
 #include "uhttpd_internal.h"
 #include "utils.h"
 #include "file.h"
-#include "ssl.h"
+
 
 static void conn_done(struct uh_connection *conn)
 {
@@ -71,7 +71,6 @@ static void conn_send(struct uh_connection *conn, const void *data, ssize_t len)
 static void conn_send_file(struct uh_connection *conn, const char *path, off_t offset, int64_t len)
 {
     struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
-    size_t min = 8192;
     struct stat st;
     int fd;
 
@@ -83,7 +82,7 @@ static void conn_send_file(struct uh_connection *conn, const char *path, off_t o
 
     fd = open(path, O_RDONLY);
     if (fd < 0) {
-        uh_log_err("open: %s\n", strerror(errno));
+        log_err("open: %s\n", strerror(errno));
         return;
     }
 
@@ -100,13 +99,21 @@ static void conn_send_file(struct uh_connection *conn, const char *path, off_t o
     if (len < 0 || len > st.st_size)
         len = st.st_size;
 
-    /* If the file is not greater than 8K, then append it to the HTTP head, send once */
-    if (len <= min) {
-        buffer_put_fd(&conni->wb, fd, len, NULL);
+    /* If the file is not greater than 2K, then append it to the HTTP head, send once */
+    if (len <= 2048) {
+        bool eof = false;
+
+        while (!eof)
+            buffer_put_fd(&conni->wb, fd, -1, &eof);
+
         close(fd);
     } else {
         conni->file.size = len;
         conni->file.fd = fd;
+#ifdef SSL_SUPPORT
+        if (conni->ssl)
+            fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0) | O_NONBLOCK);
+#endif
     }
 
     ev_io_start(conni->srv->loop, &conni->iow);
@@ -320,6 +327,29 @@ static struct uh_str conn_get_header(struct uh_connection *conn, const char *nam
     return value;
 }
 
+static void conn_traverse_headers(struct uh_connection *conn,
+        bool (*cb)(const struct uh_str name, const struct uh_str value, void *arg), void *arg)
+{
+    struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
+    struct uh_request *req = &conni->req;
+    struct uh_str name, value;
+    int i;
+
+    for (i = 0; i < UHTTPD_MAX_HEADER_NUM; i++) {
+        if (req->headers[i].field.offset == 0)
+            return;
+
+        name.p = O2D(conni, req->headers[i].field.offset);
+        name.len = req->headers[i].field.length;
+
+        value.p = O2D(conni, req->headers[i].value.offset);
+        value.len = req->headers[i].value.length;
+
+        if (!cb(name, value, arg))
+            return;
+    }
+}
+
 static uint64_t conn_get_content_length(struct uh_connection *conn)
 {
     struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
@@ -396,7 +426,7 @@ static int on_header_field_cb(struct http_parser *parser, const char *at, size_t
         req->header_num++;
 
         if (req->header_num == UHTTPD_MAX_HEADER_NUM) {
-            uh_log_err("Header too more\n");
+            log_err("Header too more\n");
             return 1;
         }
 
@@ -517,12 +547,39 @@ static struct http_parser_settings settings = {
     .on_message_complete = on_message_complete_cb
 };
 
+static void conn_incref(struct uh_connection *conn)
+{
+    struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
+
+    if (!conni)
+        return;
+
+    __sync_add_and_fetch(&conni->refcount, 1);
+}
+
+static void conn_decref(struct uh_connection *conn)
+{
+    struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
+
+    if (!conni)
+        return;
+
+    if (__sync_sub_and_fetch(&conni->refcount, 1))
+        return;
+
+    log_debug("Free connection: %p\n", conn);
+
+    free(conn);
+}
+
 void conn_free(struct uh_connection_internal *conn)
 {
     struct ev_loop *loop = conn->srv->loop;
     char addr_str[INET6_ADDRSTRLEN];
     int port;
 
+    conn->flags |= CONN_F_CLOSED;
+
     ev_timer_stop(loop, &conn->timer);
     ev_io_stop(loop, &conn->ior);
     ev_io_stop(loop, &conn->iow);
@@ -541,8 +598,8 @@ void conn_free(struct uh_connection_internal *conn)
     if (conn->next)
         conn->next->prev = conn->prev;
 
-#if UHTTPD_SSL_SUPPORT
-    uh_ssl_free(conn->ssl);
+#ifdef SSL_SUPPORT
+    ssl_session_free(conn->ssl);
 #endif
 
   if (conn->srv->conn_closed_cb)
@@ -551,12 +608,10 @@ void conn_free(struct uh_connection_internal *conn)
     if (conn->sock > 0)
         close(conn->sock);
 
-    if (uh_log_get_threshold() == LOG_DEBUG) {
-        saddr2str(&conn->addr.sa, addr_str, sizeof(addr_str), &port);
-        uh_log_debug("Connection(%s %d) closed\n", addr_str, port);
-    }
+    log_debug("Connection(%s %d) closed\n",
+            saddr2str(&conn->addr.sa, addr_str, sizeof(addr_str), &port), port);
 
-    free(conn);
+    conn_decref((struct uh_connection *)conn);
 }
 
 static void conn_http_parse(struct uh_connection_internal *conn)
@@ -600,17 +655,49 @@ static void conn_http_parse(struct uh_connection_internal *conn)
     }
 }
 
-#if UHTTPD_SSL_SUPPORT
-static int conn_ssl_write(int fd, void *buf, size_t count, void *ssl)
+#ifdef SSL_SUPPORT
+static void on_ssl_verify_error(int error, const char *str, void *arg)
 {
-    int ret = uh_ssl_write(ssl, buf, count);
-    if (ret < 0) {
-        if (ret == UH_SSL_ERROR_AGAIN)
-            return P_FD_PENDING;
+    log_warn("SSL certificate error(%d): %s\n", error, str);
+}
+
+/* -1 error, 0 pending, 1 ok */
+static int ssl_negotiated(struct uh_connection_internal *conn)
+{
+    char err_buf[128];
+    int ret;
+
+    ret = ssl_connect(conn->ssl, true, on_ssl_verify_error, NULL);
+    if (ret == SSL_PENDING)
+        return 0;
+
+    if (ret == SSL_ERROR) {
+        log_err("ssl connect error(%d): %s\n", ssl_err_code, ssl_strerror(ssl_err_code, err_buf, sizeof(err_buf)));
+        return -1;
+    }
+
+    conn->flags &= CONN_F_SSL_HANDSHAKE_DONE;
+
+    return 1;
+}
+
+static int conn_ssl_read(int fd, void *buf, size_t count, void *arg)
+{
+    struct uh_connection_internal *conn = arg;
+    static char err_buf[128];
+    int ret;
+
+    ret = ssl_read(conn->ssl, buf, count);
+    if (ret == SSL_ERROR) {
+        log_err("ssl_read(%d): %s\n", ssl_err_code,
+                ssl_strerror(ssl_err_code, err_buf, sizeof(err_buf)));
         return P_FD_ERR;
     }
-    return ret;
 
+    if (ret == SSL_PENDING)
+        return P_FD_PENDING;
+
+    return ret;
 }
 #endif
 
@@ -619,41 +706,74 @@ static void conn_write_cb(struct ev_loop *loop, struct ev_io *w, int revents)
     struct uh_connection_internal *conn = container_of(w, struct uh_connection_internal, iow);
     int ret;
 
-#if UHTTPD_SSL_SUPPORT
-    if (conn->ssl)
-        ret = buffer_pull_to_fd_ex(&conn->wb, w->fd, buffer_length(&conn->wb), conn_ssl_write, conn->ssl);
-    else
-#endif
-        ret = buffer_pull_to_fd(&conn->wb, w->fd, buffer_length(&conn->wb));
+    if (conn->ssl) {
+#ifdef SSL_SUPPORT
+        static char err_buf[128];
+        struct buffer *b = &conn->wb;
 
-    if (ret < 0) {
-        uh_log_err("write error: %s\n", strerror(errno));
-        conn_free(conn);
-        return;
+        if (!likely((conn->flags & CONN_F_SSL_HANDSHAKE_DONE))) {
+            ret = ssl_negotiated(conn);
+            if (ret < 0)
+                goto err;
+            if (ret == 0)
+                return;
+        }
+
+        ret = ssl_write(conn->ssl, buffer_data(b), buffer_length(b));
+        if (ret == SSL_ERROR) {
+            log_err("ssl_write(%d): %s\n", ssl_err_code,
+                    ssl_strerror(ssl_err_code, err_buf, sizeof(err_buf)));
+            goto err;
+        }
+
+        if (ret == SSL_PENDING)
+            return;
+
+        buffer_pull(b, NULL, ret);
+#endif
+    } else {
+        ret = buffer_pull_to_fd(&conn->wb, w->fd, -1);
+        if (ret < 0) {
+            log_err("write error: %s\n", strerror(errno));
+            goto err;
+        }
     }
 
     if (buffer_length(&conn->wb) == 0) {
         if (conn->file.fd > 0) {
-            ret = sendfile(w->fd, conn->file.fd, NULL, conn->file.size);
-            if (ret < 0) {
-                if (errno != EAGAIN) {
-                    uh_log_err("write error: %s\n", strerror(errno));
-                    conn_free(conn);
+#ifdef SSL_SUPPORT
+            if (conn->ssl) {
+                bool eof;
+                if (buffer_put_fd(&conn->wb, conn->file.fd, 8192, &eof) < 0 || eof) {
+                    close(conn->file.fd);
+                    conn->file.fd = -1;
                 }
                 return;
-            }
+            } else {
+#endif
+                ret = sendfile(w->fd, conn->file.fd, NULL, conn->file.size);
+                if (ret < 0) {
+                    if (errno != EAGAIN) {
+                        log_err("write error: %s\n", strerror(errno));
+                        goto err;
+                    }
+                    return;
+                }
 
-            if (ret < conn->file.size) {
-                conn->file.size -= ret;
-                return;
-            }
+                if (ret < conn->file.size) {
+                    conn->file.size -= ret;
+                    return;
+                }
 
-            close(conn->file.fd);
-            conn->file.fd = -1;
+                close(conn->file.fd);
+                conn->file.fd = -1;
+#ifdef SSL_SUPPORT
+            }
+#endif
         }
 
         if (conn->flags & CONN_F_SEND_AND_CLOSE) {
-            conn_free(conn);
+            goto err;
         } else {
             ev_io_stop(loop, w);
 
@@ -663,20 +783,12 @@ static void conn_write_cb(struct ev_loop *loop, struct ev_io *w, int revents)
                 conn_http_parse(conn);
         }
     }
-}
 
-#if UHTTPD_SSL_SUPPORT
-static int conn_ssl_read(int fd, void *buf, size_t count, void *ssl)
-{
-    int ret = uh_ssl_read(ssl, buf, count);
-    if (ret < 0) {
-        if (ret == UH_SSL_ERROR_AGAIN)
-            return P_FD_PENDING;
-        return P_FD_ERR;
-    }
-    return ret;
+    return;
+
+err:
+    conn_free(conn);
 }
-#endif
 
 static void conn_read_cb(struct ev_loop *loop, struct ev_io *w, int revents)
 {
@@ -690,41 +802,38 @@ static void conn_read_cb(struct ev_loop *loop, struct ev_io *w, int revents)
         return;
     }
 
-#if UHTTPD_SSL_SUPPORT
-    if (conn->ssl && !(conn->flags & CONN_F_SSL_HANDSHAKE_DONE)) {
-        ret = uh_ssl_handshake(conn->ssl);
-        if (ret == UH_SSL_ERROR_AGAIN)
-            return;
-        if (ret == UH_SSL_ERROR_UNKNOWN) {
-            conn_free(conn);
-            return;
-        }
-        conn->flags |= CONN_F_SSL_HANDSHAKE_DONE;
-    }
-#endif
-
     conn->activity = ev_now(loop);
 
-#if UHTTPD_SSL_SUPPORT
-    if (conn->ssl)
-        ret = buffer_put_fd_ex(rb, w->fd, -1, &eof, conn_ssl_read, conn->ssl);
-    else
-#endif
-        ret = buffer_put_fd(rb, w->fd, -1, &eof);
+    if (conn->ssl) {
+#ifdef SSL_SUPPORT
+        if (!likely((conn->flags & CONN_F_SSL_HANDSHAKE_DONE))) {
+            ret = ssl_negotiated(conn);
+            if (ret < 0)
+                goto err;
+            if (ret == 0)
+                return;
+        }
 
-    if (ret < 0) {
-        uh_log_err("read error: %s\n", strerror(errno));
-        goto done;
+        ret = buffer_put_fd_ex(&conn->rb, w->fd, -1, &eof, conn_ssl_read, conn);
+        if (ret < 0)
+            goto err;
+#endif
+    } else {
+        ret = buffer_put_fd(rb, w->fd, -1, &eof);
+        if (ret < 0) {
+            log_err("read error: %s\n", strerror(errno));
+            goto err;
+        }
     }
 
     if (eof)
-        goto done;
+        goto err;
 
     conn_http_parse(conn);
 
     return;
 
-done:
+err:
     conn_free(conn);
 }
 
@@ -787,20 +896,25 @@ static void conn_init_cb(struct uh_connection *conn)
     conn->get_path = conn_get_path;
     conn->get_query = conn_get_query;
     conn->get_header = conn_get_header;
+    conn->traverse_headers = conn_traverse_headers;
     conn->get_content_length = conn_get_content_length;
     conn->get_body = conn_get_body;
     conn->extract_body = conn_extract_body;
 
     conn->close = conn_close;
+
+    conn->incref = conn_incref;
+    conn->decref = conn_decref;
 }
 
-struct uh_connection_internal *uh_new_connection(struct uh_server_internal *srv, int sock, struct sockaddr *addr)
+struct uh_connection_internal *uh_new_connection(struct uh_listener *l, int sock, struct sockaddr *addr)
 {
+    struct uh_server_internal *srv = l->srv;
     struct uh_connection_internal *conn;
 
     conn = calloc(1, sizeof(struct uh_connection_internal));
     if (!conn) {
-        uh_log_err("malloc: %s\n", strerror(errno));
+        log_err("malloc: %s\n", strerror(errno));
         return NULL;
     }
 
@@ -821,9 +935,9 @@ struct uh_connection_internal *uh_new_connection(struct uh_server_internal *srv,
     ev_timer_init(&conn->timer, keepalive_cb, UHTTPD_CONNECTION_TIMEOUT, 0.0);
     ev_timer_start(srv->loop, &conn->timer);
 
-#if UHTTPD_SSL_SUPPORT
-    if (srv->ssl_ctx)
-        conn->ssl = uh_ssl_new(srv->ssl_ctx, sock);
+#ifdef SSL_SUPPORT
+    if (l->ssl)
+        conn->ssl = ssl_session_new(srv->ssl_ctx, sock);
 #endif
 
     http_parser_init(&conn->parser, HTTP_REQUEST);
@@ -832,5 +946,9 @@ struct uh_connection_internal *uh_new_connection(struct uh_server_internal *srv,
 
     conn_init_cb(&conn->com);
 
+    conn_incref((struct uh_connection *)conn);
+
+    log_debug("New connection: %p\n", conn);
+
     return conn;
 }
